<?php
/**
 * @file API examples for the import_html hooks
 *
 * Examples of all these hooks are in the extras directory. The hooks were
 * developed as needed as the tasks required by the 'extras' were identified.
 *
 * As well as module hooks, "Rules" actions of the same name are triggered
 * at the same points. Hooks first, then Actions.
 */

/**
 * Current phases:
 *
 * * import_html_pre_import_file :
 *     Only the source filename is available. Can initialize global settings.
 * * import_html_preprocess :
 *     The raw source is available. Can run string-only repairs on it.
 * * import_html_preparse :
 *     The DOM is loaded, but not yet transformed. Perform XML cleanups on
 *     $source_node->xmldoc but not data extractions.
 * * import_html :
 *     The core process. All the data can be extracted from the xml doc
 *     and stored into a scratch $node object.
 * * import_html_merge :
 *     Deals with issues around merging new data onto an existing entity
 * * import_html_after_save :
 *     Triggered to perform action that are only valid after an entity ID is
 *     created.
 */

/**
 * Definition of the import_html $profile or $context array.
 *
 * An import process starts with a $profile, which is the main repository for
 * settings, and is passed around as the director of the current import task.
 * Then rules and internal logic may add a few attributes to it, at which point
 * it has become the $context array.
 * $profile settings plus runtime parameters = $context.
 *
 * Many of the settings are defined in the function
 * import_html_profile_defaults(), are visible on the profile settings page,
 * and set the basic rules for importing.
 * These rules can be changed on the fly by later processes, eg if a jpeg
 * is detected, the path to save the download in may be changed.
 *
 * Additional settings are exposed by the add-on extras. They should describe
 * themselves in their own hook_import_html_profile_defaults.
 *
 *
 * Some further flags or values that may be set or used during processing or
 * 'runtime' for the $context.
 * The primary one is 'rel_path', which defines the current working path.
 *
 * At runtime, some other internal flags include :
 *
 * * 'import_html_exclude' which is set if the page is to be ignored.
 *
 * * 'you_shall_not_parse' is used to indicate binary files or files to be
 *    left as-is. They will not become managed entities. Normally this
 *    is the case for all non-html MIME types.
 *
 * * 'dry_run' is a context flag for internal use only. It's used by the
 *    preview and demo wizards to ensure that no data gets saved, even though
 *    a full import process normally requires a placeholder node to be created
 *    and then updated. Even when using dry_run however, downloads may happen.
 *
 * Later data - actual content - gets set on the working copy of the $node
 * object. $context should only hold on to info about the current task.
 */

/**
 * This gets called for each file request, before any data is retrieved.
 *
 * It's most often used to decide - based on pathname alone - if a file should
 * be skipped, or treated as binary, or have a different sort of profile
 * used.
 *
 * Now is the only time to change the $context['dest_path'] also.
 *
 * This can also be used for general setup of preferences.
 *
 * The rel_path will be available as a key within the $context.
 *
 * @param array $context
 *   A loaded profile, with some additions for the current process.
 */
function hook_import_html_pre_import_file(array &$context) {
  if (import_html_guess_file_class($context['rel_path']) == 'document') {
    $context['import_html_exclude'] = TRUE;
  }
}


/**
 * This gets called when the raw text data is first retrieved, before it's
 * read into the DOM.
 * Use this if very bad markup needs cleaning.
 *
 * @param array $profile
 * @param string $raw_source
 *   Modify the full string by reference.
 */
function hook_import_html_preprocess(array &$profile, &$raw_source) {
  $raw_source = preg_replace('@<[^>]*name="__VIEWSTATE"[^>]*>@', '', $raw_source);
}

/**
 * The DOM is loaded, but the XSL template has not yet been run.
 * Now is a good time for link rewriting and removing unwanted elements with
 * XML methods.
 *
 * The thing to act on here is the $node->xmldoc.
 * Attributes set on the $node will be kept through the process, though
 * may be overwritten.
 *
 * @param array $profile
 * @param string $context
 * @param object $node
 */
function hook_import_html_preparse(array &$context, object &$node) {
   // The thing to act on here is the $node->xmldoc.

   // Unpack all spans, but don't discard their content
  $found_elements = xml_getElementsByTagName($node->xmldoc, 'span');
  foreach ($found_elements as $found_element) {
    $parent = $found_element->parentNode;
    $where_to_insert_before = $found_element;
    while ($lastchild = $found_element->lastChild) {
      $val = trim(xml_toString($lastchild));
      if (empty($val)) {
        $found_element->removeChild($lastchild);
        continue;
      }
      $where_to_insert_before = $parent->insertBefore($lastchild, $where_to_insert_before);
    }
    $found_element->parentNode->removeChild($found_element);
  }
}

/**
 * HOOK_import_html is the main callback function.
 *
 * It runs after all the basic parsing has been done, when we have most of a
 * fully-formed node object ready to save.
 *
 * The $node object will contain a large amount of extra attributes - values
 * that were added during the semantic data extraction. Most of the supported
 * modules that add their own data to the $node run at this point. It's like a
 * nodeapi presave function. They will run simultaneously with your modules
 * implementation, according to module weight.
 *
 * $node->import_html_exclude
 * $node->import_html_no_menus
 *
 * At this point in the process, we never know a $node->nid value.
 *
 * @param $profile An array of settings, containing the rules for this import
 * process.
 *
 * @param $node The Node object. Many non-standard attributes are attached to
 * this proto-object. Some extra flags (see below) can be set on it to influence
 * later save behavior.
 *
 * @param @datadoc An XML DOM object, containing the transformed (not raw)
 * version of the source doc.
 */
function hook_import_html(array &$context, object &$node, XMLDomElement &$datadoc = NULL) {

  // FLAG
  // $node->import_html_exclude
  // will prevent this node from being saved altogether.
  // It can be set to a useful text string that will go into the log,
  // but it just needs to be set to anything non-null


  // EXAMPLE RULE
  // Prevent certain paths from being processed
  $excluded_paths = array(
    'docs',
    'apps',
    'archive',
  );
  foreach ($excluded_paths as $excluded_path) {
    if (preg_match('|' . $excluded_path . '|', $node->path_alias) ) {
      $node->import_html_exclude = "Matched exclusion pattern '$excluded_path' ";
      import_html_debug("Skipping the import of this page as it matches the exclusion pattern %pattern", array('%pattern' => $excluded_path), WATCHDOG_NOTICE);
    }
  }

  // FLAG
  // $node->import_html_no_menus
  // can be set to stop menu.module from adding a menu item.
  // This overrides the profile default.

  // Scan the document for any class that was set on the body element.
  $body_elements = xml_query($datadoc, '//xhtml:body[@class]');
  foreach ($body_elements as $body_element) {
    $classes = explode(' ', xml_getattribute($body_element, 'class'));
  }

  // EXAMPLE RULE
  // FAQ pages don't get a menu item
  if (in_array('faq', $classes)) {
    $node->import_html_no_menus = "FAQ pages get no menu item";
  }

  // EXAMPLE RULE
  // Pages of class 'product' need to become a 'product' node type
  if (in_array('product', $classes)) {
    $node->type = "product";
  }

}

/**
 * HOOK_import_html_merge runs when an existing node is being overwritten.
 *
 * It can be used to selectively copy content across or to avoid conflicts. By
 * default, conflicting arrays of data will be merged, and conflicting strings
 * will be overwritten
 *
 * This func is only called when a valid old_node was found, and is not called
 * if none exists.
 */
function hook_import_html_merge($profile, &$node, $old_node) {
  // EXAMPLE RULE
  // If the old node already has a title,
  // do not overwrite it with the imported version.
  // The simplest way to ensure that is to set the new node title to the old node title.
  $node->title = $old_node->title;
}


/**
 * HOOK_import_html_after_save runs after the node has been created.
 *
 * Some updates can only happen after the node has been saved and we know the
 * new nid.
 *
 * @see menu_import_html_after_save
 */
function hook_import_html_after_save($profile, &$node, $datadoc = NULL) {

}