<?php

/**
 * Publish our actions and logic to rules API
 */


/**
 * Publish that "importing a page" is an action that rules can act on.
 *
 * Events to fire are:
 *
 *   pre_import_file - where the context and the source is evaluated to decide
 * which import profile to follow.
 * No data is available at this time. Choices made here are based on path
 * patterns only (is it a binary, is it under /news etc)
 *
 *   process_data -
 * Inital naive data scans (the raw text, and DOM) may be available here, but
 * is read-only.
 *
 *   import_page - when the import profile has been chosen and the data is now
 * being built.
 *
 *   post_import_page - when most of the data has been imported and the new
 * node saved. Now you can do the extra data work that required a new entity ID.
 *
 * @see HOOK_rules_action_info()
 */

function import_html_rules_event_info() {
  $items = array(
    'import_html_pre_import_file' => array(
      'label' => t('Before importing a file'),
      'group' => t('Import'),
      'help' => t("
        When this is fired, the source data has not yet been fetched,
        in fact you may decide not to. Only the pathname is really available
        right now. Use this to choose different profiles based on
        pathname or filetype.
      "),
      'variables' => array(
        'profile' => array(
          'type' => 'import_html_profile',
          'label' => t('Page import profile'),
          'description' => t('Import_HTML settings currently in effect'),
        ),
        'context' => array(
          'type' => 'import_html_context',
          'label' => t('Task context'),
          'description' => t('
            Current context about the path and data to be processed.
            Parsing has not yet been done.
          '),
        ),
      ),
    ),

    'import_html_process_data' => array(
      'label' => t('When processing data to import'),
      'group' => t('Import'),
      'help' => t("
        When this is fired, the source data has been fetched and can be
        inspected for values.
        The target entity is not being built yet, use this step to choose
        settings based on read-only values you can extract from the import
        data. eg if generator='otherCMS' then profile.contentID='#main'.
      "),
      'variables' => array(
        'profile' => array(
          'type' => 'array',
          'label' => t('Site import profile')
        ),
        'context' => array(
          'type' => 'import_html_context',
          'label' => t('task context')
        ),
        'source_data' => array(
          'type' => 'import_html_source_data',
          'label' => t('source data')
        ),
      ),
    ),

    'import_html_import_page' => array(
      'label' => t('When importing a page'),
      'group' => t('Import'),
      'help' => t("
        The data has been prepared,
        the import profile settings are pretty much decided,
        the field mappings are defined,
        a target entity is being built.
        Use this step tp set other values on the target as needed.
      "),
      'variables' => array(
        'profile' => array(
          'type' => 'array',
          'label' => t('Site import profile')
        ),
        'context' => array(
          'type' => 'import_html_context',
          'label' => t('task context')
        ),
        'source_data' => array(
          'type' => 'import_html_source_data',
          'label' => t('source data')
        ),
        'entity' => array(
          'type' => 'entity',
          'label' => t('Target entity')
        ),
      ),
    ),

    'import_html_post_import_page' => array(
      'label' => t('After importing a page'),
      'group' => t('Import'),
      'help' => t("
        Most of the data has been imported and the new entity is saved.
        Now you can do the extra data work that required a new entity ID.
      "),
      'variables' => array(
        'profile' => array(
          'type' => 'array',
          'label' => t('Site import profile')
        ),
        'context' => array(
          'type' => 'import_html_context',
          'label' => t('task context')
        ),
        'entity' => array(
          'type' => 'entity',
          'label' => t('Target entity')
        ),
      ),
    ),
  );
  return $items;
}

/**
 * Declare that our profile settings can be exposed to use as arguments within
 * rules.
 * This, plus the schem defined in import_html_entity_info() means the Feeds UI
 * will help to expose the attributes that are set on an import_html profile
 * or task context.
 *
 *  @see rules_rules_core_data_info()
 */
function import_html_rules_data_info() {
  $import_html_profile = entity_get_info('import_html_profile');
  $return = array(
    'import_html_profile' => array(
      'label' => t('Import_HTML profile settings'),
      'group' => t('Import'),
      'wrap' => TRUE,
#      'property info' => $import_html_profile['property info'],
    ),

    'import_html_context' => array(
      'label' => t('Import_HTML current source info'),
      'group' => t('Import'),
      'wrap' => TRUE,
      'property info' => array(
        'profile_id' => array('type' => 'text', 'label' => 'Profile ID'),
        'rel_path' => array('type' => 'text', 'label' => 'Relative path'),
      ),
    ),

    'import_html_source_data' => array(
      'label' => t('Import_HTML data being scanned'),
      // In import_html terms, this is the $source_node where data accumulates
      'group' => t('Import'),
      'wrap' => TRUE,
      'property info' => array(
        'raw' => array('type' => 'text', 'label' => 'Raw text'),
        'dom' => array('type' => 'object', 'label' => 'Parsed DOM'),
        'data' => array(
          'type' => 'array',
          'label' => 'Extracted values',
          'description' => '
            A number of key-value pairs automatically extracted
            from the markup semantics, or built by previous rules.
          ',
        ),
      ),
    ),

  );
dpm($return);
  return $return;
}



/**
 * Advertise the sorts of conditions we look for.
 *
 * Array keys must match callback functions.
 *
 * @see hook_rules_condition_info()
 */
function import_html_rules_condition_info() {
  return array(
    'import_html_condition_path_compare' => array(
      'label' => t('Compare the relative path being processed against a pattern'),
      'parameter' => array(
        'context' => array(
          'label' => t('Current task context'),
          'type' => 'import_html_context',
        ),
        'pattern' => array(
          'label' => t('Pattern'),
          'type' => 'text'
        ),
      ),
      'group' => t('Import'),
    ),

    'import_html_condition_suffix_compare' => array(
      'label' => t('Compare the file being processed against known suffixes'),
      'parameter' => array(
        'context' => array(
          'label' => t('Current task context'),
          'type' => 'import_html_context',
        ),
        'pattern' => array(
          'label' => t('Pattern'),
          'type' => 'text'
        ),
      ),
      'group' => t('Import'),
    ),

    'import_html_condition_data_pattern_scan' => array(
      'label' => t('Data being imported matches pattern'),
      'parameter' => array(
        'context' => array(
          'label' => t('Current data'),
          'type' => 'import_html_data',
        ),
        'pattern' => array(
          'label' => t('Pattern'),
          'type' => 'text',
        ),
        'pattern_type' => array(
          'label' => t('Pattern type'),
          'type' => 'select',
        ),
      ),
      'group' => t('Import'),
    ),
  );
}


/**
 * Condition: path matches the pattern
 */
function import_html_condition_path_compare($import_html_context, $pattern) {
  return TRUE;
}

/**
 * Condition: suffix matches the pattern
 */
function import_html_condition_suffix_compare($import_html_context, $pattern) {
  return TRUE;
}

/**
 * Condition: data matches the pattern
 */
function import_html_condition_data_pattern_scan($import_html_context, $pattern) {
  return TRUE;
}


/**
 * Define the actions that a rule can trigger.
 *
 * Implements of hook_rules_action_info().
 */
function import_html_rules_action_info() {
  return array(
    'import_html_action_skip_item' => array(
      'label' => t('Skip import of item'),
      'group' => t('Import'),
      'parameter' => array(
        'entity' => array(
          'type' => 'entity',
          'label' => t('The import item to be marked as skipped')
        ),
      ),
    ),

    'import_html_action_set_profile' => array(
      'label' => t('Set the import_html profile to use'),
      'group' => t('Import'),
      'parameter' => array(
        'profile' => array(
          'type' => 'list<text>',
          'label' => t('The import_html settings profile to use'),
          'options list' => 'import_html_profile_list',
        ),
      ),
    ),

    'import_html_action_set_target_entity_type' => array(
      'label' => t('Set the target content type to save into'),
      'group' => t('Import'),
      'parameter' => array(
        'profile' => array(
          'type' => 'list<text>',
          'label' => t('Available content types'),
          'options list' => 'node_type_get_names',
        ),
      ),
    ),

  );
}


/**
 * Mark import item as skipped.
 * @see import_html_rules_action_info()
 */
function import_html_action_skip_item($entity_wrapper) {
  $entity = $entity_wrapper->value();
  if (isset($entity->feeds_item)) {
    $entity->feeds_item->skip = TRUE;
  }
}

/**
 * Change the import_profile settings being used for this import process.
 * @see import_html_rules_action_info()
 */
function import_html_action_set_profile($entity_wrapper, $profile) {
  $entity = $entity_wrapper->value();
}

/**
 * Change the import_profile settings being used for this import process.
 * @see import_html_rules_action_info()
 */
function import_html_action_set_target_entity_type($entity_wrapper, $profile) {
  $entity = $entity_wrapper->value();
}

